MINIO_ETCD_ENDPOINTS="http://localhost:2379" GOOGLE_CLOUD_CREDENTIALS=~/.cloud/gcloud/application_default_credentials.json MINIO_ACCESS_KEY=vanpelt MINIO_SECRET_KEY=fde21a1af035c676807711c4d60ca6f02a3689de minio gateway gcs wandb-production
